#!/usr/bin/perl
#
#    nmis.pl - NMIS Perl Program - Network Mangement Information System
#    Copyright (C) 2000 Sinclair InterNetworking Services Pty Ltd
#    Keith Sinclair <keith@sinclair.org.au>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#*****************************************************************************
# Auto configure to the <nmis-base>/lib and <nmis-base>/files/nmis.conf
use FindBin;
use lib "$FindBin::Bin/../lib";
use lib "/usr/local/rrdtool/lib/perl";

#
#****** Shouldn't be anything else to customise below here *******************
# best to customise in the nmis.conf file.
#

# Check if there are any arguements
#! New commandline mthread=true for multithreading and mthreaddebug=true for debuging only the threading process and
#! the interprocess communication
if ( $#ARGV < 0 ) {
	print <<EO_TEXT;
$0 NMIS Polling Engine - Network Mangement Information System
Copyright (C) 2000 Sinclair InterNetworking Services Pty Ltd

command line options are:
    type=<option>             Where <option> is one of the following:
                                   collect   NMIS will collect all statistics;
                                   update    Update all the dynamic NMIS configuration
                                   threshold Check collected stats for thresholds
                                   config    Validate the chosen configuration file
                                   apache    Produce Apache configuration for NMIS
    [file=<file name>]             Optional alternate configuation files
    [node=<node name>]             Run operations on a single node
    [debug=true|false|verbose|0-9] default=false - Show debuging information, quite handy
    [collect=true|false]           default=true - Do the SNMP collection or not
    [mthread=true|false]           default=false - Enable Multithreading or not
    [mthreaddebug=true|false]      default=false - Enable Multithreading debug or not
    [maxthreads=<1..XX>]           default=2 - How many threads should nmis create
EO_TEXT
	exit(0);
}

require 5;

use Time::HiRes;
my $startTime = Time::HiRes::time();


use strict;
use csv;
use BER;
use SNMP_Session;
use SNMP_MIB;
use SNMP_Simple;
use SNMPv2c_Simple;
use RRDs;
use NMIS;
use func;
use ip;
use ping;

# Global Interpreted Variables ie stuff we figured out.
my $sysObjectName;
my $ifIndex;
my $ifSpeed;

# Variable used to control writing the system file to ensure that it doesn't get blanked.
my $writeSystem = "false";

# Time Variables
my $time;
my $datestamp;
my $datestamp_2;
my $endTime;
# Global SNMP Variables once we get em keep em global
my %snmpTable;

# Variables for command line munging
my %nvp = getArguements(@ARGV);

# See if customised config file required.
my $conf;
if ( $nvp{file} ne "" ) { $conf = $nvp{file}; }
else { $conf = "nmis.conf"; }
my $configfile = "$FindBin::Bin/../conf/$conf";
if ( -f $configfile ) { loadConfiguration($configfile); }
else { die "Can't access configuration file $configfile.\n"; }

#Load the configuration
loadConfiguration($configfile);

# all arguments are now stored in nvp (name value pairs)
my $type=$nvp{type};
my $graph=$nvp{graph};
my $collect=$nvp{collect};
my $node_search=$nvp{node};
my $rmefile=$nvp{rmefile};
#! store multithreading arguments in nvp
my $mthread=($nvp{mthread} eq "true") ? 1 :0;
my $mthreadDebug=($nvp{mthreaddebug} eq "true") ? 1 :0;
my $maxThreads=$nvp{maxthreads};

# Set debugging level.
my $debug = setDebug($nvp{debug});
$NMIS::debug = $debug;

# Create a hash for storing runReachability stats
my %reach;

# Create a hash for storing interface stats
my %ifStats;

# An array of Interface Status for AdminStatus and OperStatus
my @interfaceStatus = ("null","up","down","testing","unknown","dormant","notPresent");

# Set the default collect option
if ($collect eq "") { $collect = "true"; }

# Set the default graph option
if ($graph eq "") { $graph = "false"; }

#! Set the default maxThread option. This controlls the maximum numbers of threads nmis will create.
if ($maxThreads eq "") { $maxThreads = 2;}

# Other Global Variables
my $database;
my $databaseOriginal = "";
my @interfaceTable;
my $tmpifDescr;
my $ping;
my $timeout = 10;
my $pingresult;
my $snmpresult;
my $run;
my $statsDir;
my $devicetype;
my $onenode;
my $index;
my $nodecount;

# The node to turn on debugging for no matter what
my $debug_watch = "";
my $debug_global = $debug;

#! ******************************************************************
#! **         multithreading variables and IPC stuff
#! ******************************************************************
#use POSIX ":sys_wait_h";
#! using the interprocess shareable.pm (can be downloaded from www.cpan.org)
#use IPC::Shareable;
#! Use the PROC::Queue library from Salvador Garcia. Can be downloaded from http://search.cpan.org/search?module=Proc::Queue
#! This library does all the IPC stuff, so we mustn't do it. Therefore we don't need IPC::Shareable anymore
use Proc::Queue;
$SIG{CHLD} = 'IGNORE';
#! if your hit ctrl+c druring multithreading, the childs will die and produces some output
$SIG{INT} = sub { die "Child with PID $$ is dying\n"};
my $currentThreads;
#! var for the IPC communication
#my $handle;
#! creating a golbal variable in shared memory. Alle processes can access this variable (even the childs). See IPC::Shareable documentation.
#$handle = tie $currentThreads, 'IPC::Shareable', undef, { destroy => 1 };
#! sets the maximum number of forked childs
Proc::Queue::size($maxThreads);
#! debug the forking process
Proc::Queue::debug($mthreadDebug);
#! delay lets you set a minimun time in seconds to elapse between every consecutive calls
#!  to fork. This is usefull for not creating to much processes in a very short time.
Proc::Queue::delay(0.2);
my $pid;

# How long to run for in seconds thus 30 * 60 is 1800 seconds
my $start_time = time;
my $run_time = 60 * 60;

if ($debug) { print returnTime ." Starting $0 $NMIS::VERSION: type=$type file=$configfile node=$node_search run_time=$run_time\n"; }

# Load the configuration and the node file!!!!!!!
loadConfiguration($configfile);
loadNodeDetails;
	
##############################################################
# SNMP Setup, debug options, load base mibs etc.
##############################################################
# set debug on or not depending on global debug 
if ($debug) {
	#You can turn off SNMP debug but still leave debug 
	#messages on just comment the following line for no 
	#SNMP debug at all
	#$SNMP_Simple::debug=1;
	
	# If you want SNMP warnings comment these lines
	#$SNMP_Simple::suppress_warnings=1;
	#$SNMP_Session::suppress_warnings=1;
}
else {
	$SNMP_Simple::suppress_warnings=1;
	$SNMP_Session::suppress_warnings=1;
	$SNMP_MIB::suppress_warnings=1;
}

# initalizes the OID mappings for the SNMP_Simple library.
if ($debug) { print returnTime." Loading All the Mibs\n"; }

if ($debug > 1) { print "\tLoading $NMIS::config{mib_root}/$NMIS::config{standard_mib}\n"; }
if ( -f "$NMIS::config{mib_root}/$NMIS::config{standard_mib}" ) { 
	SNMP_MIB::loadoids_file("$NMIS::config{mib_root}", "$NMIS::config{standard_mib}"); 
} else { warn returnTime." nmis.pl, mib file $NMIS::config{mib_root}/$NMIS::config{standard_mib} not found.\n"; }	

if ($debug > 1) { print "\tLoading $NMIS::config{mib_root}/$NMIS::config{router_mib}\n"; }
if ( -f "$NMIS::config{mib_root}/$NMIS::config{router_mib}" ) { 
	SNMP_MIB::loadoids_file("$NMIS::config{mib_root}", "$NMIS::config{router_mib}"); 
} else { warn returnTime." nmis.pl, mib file $NMIS::config{mib_root}/$NMIS::config{router_mib} not found.\n"; }	

if ($debug > 1) { print "\tLoading $NMIS::config{mib_root}/$NMIS::config{switch_mib}\n"; }
if ( -f "$NMIS::config{mib_root}/$NMIS::config{switch_mib}" ) { 
	SNMP_MIB::loadoids_file("$NMIS::config{mib_root}", "$NMIS::config{switch_mib}");
} else { warn returnTime." nmis.pl, mib file $NMIS::config{mib_root}/$NMIS::config{switch_mib} not found.\n"; }

# If this is an update, check the config and report any errors
if ( $type eq "update" ) { &checkConfig; }

# If this is an escalation run, do it
if ( $type eq "escalate" ) { &runEscalate; }
# If this is an config run, do it
elsif ( $type eq "config" ) { &checkConfig("fix"); }
# If this is an apache run, do it
elsif ( $type eq "apache" ) { &printApache; }
# If this is a threshold run, do it
elsif ( $type eq "threshold" )
{
	#######################################################################
	# Do the threshold thing
	#######################################################################
	if ( $node_search eq "" )
	{
		#! multithreading
		#! currentthreads should be 0....
		if ($mthread)
		{
			#! shlock will lock the shared memory var $currentThreads so that no other process will write something in it.
			#! all other processes will be queued! So nothing will get lost.
			#$handle->shlock();
			#$currentThreads=0;
			#! unlocks the var
			#$handle->shunlock();
		}

		foreach $onenode ( keys (%NMIS::nodeTable) )
		{
			#! One thread for each node until maxThreads is reached.		
			#! This loop is entered only if the commandlinevariable mthread=true is used!
  	  if ($mthread)
			{
				#! incrementing the currentThread value for each thread which is startet. This incrementation will be done in the
				#! father process because it can take a time until the child is borne and running
				#!$handle->shlock();
				#!$currentThreads++;
				#!$handle->shunlock();
			
				if ($mthreadDebug)
				{
	      	print "FATHER-> running threshold in multithreading modus\n";
				}

				#! starting new child process
      	if ($pid=fork)
				{
					#! this will be run only by the father process
					if ($mthreadDebug)
					{
						print "FATHER-> starting a thread for node $onenode\n";
						$currentThreads=Proc::Queue::running_now();
						print "FATHER-> currently there are $currentThreads threads running\n";
					}
				}

				#! this will be run only by the child
				else
				{
					if ($mthreadDebug)
					{				
						print "CHILD $$-> I am a CHILD with the PID $$ \n";
						print "CHILD $$-> My father's PID is ";print getppid();print "\n";
					}

					&runThreshold($onenode);

				  #! all the work in this thread is done
					#! now this child will die. But before it must decrement the currentThreads var
					if ($mthreadDebug)
					{				
						$currentThreads=Proc::Queue::running_now();
						print "CHILD $$-> currently there are $currentThreads running\n";
					}

					#!$handle->shlock();
					#!$currentThreads=$currentThreads-1;
					#!$handle->shunlock();

					if ($mthreadDebug)
					{				
						$currentThreads=Proc::Queue::running_now();
						print "CHILD $$-> now there are $currentThreads running\n";
						print "CHILD $$-> I WiLl DiE nOw aRgHhHhHhHhH\n";
					}
				
					#! killing child
					exit 0;
				}
			}
			else
			{
				&runThreshold($onenode);
			}

			#! Check how much threads are running.
#!			if ($mthread)
#!			{	
#!				if ($currentThreads == $maxThreads)
#!				{
#!					if ($mthreadDebug)
#!					{
#!						print "FATHER-> maximum number of concurrent threads is reached!\n";
#!						print "FATHER-> there are $currentThreads  Threads\n";
#!					}
				
					#! wait until one or more childs are dead....
#!					while ($currentThreads >= $maxThreads)
#!					{
#!						if ($mthreadDebug)
#!						{
#!							print "FATHER-> must wait until one or more childs are dead. Currently there are $currentThreads Threads\n";
#!						}
					
						#! sleep a while. is good for the cpu....
#!						select(undef, undef, undef, 0.3);
#!					}

#!					if ($mthreadDebug)
#!					{
#!						print "FATHER-> One or more Childs are dead now. So I can fork some new ones.\n";
#!						print "FATHER-> currently there are $currentThreads Threads\n";
#!					}
#!				}
#!			} #if mthread
		}#! foreach
	}#! if ($node_search eq)
	else
	{
		&runThreshold($node_search);
	}
} #!elsif threshold

# Read a Cisco Works file and produce an NMIS Nodes file
elsif ( $type eq "rme" ) {
	if ( $rmefile eq "" ) {
		errorQuit ("$0 the type=rme option requires a file arguement for source rme CSV file\ni.e. $0 type=rme rmefile=/data/file/rme.csv");
	}
	else {
		loadRMENodes($rmefile);
		writeNodesFile("$NMIS::config{Nodes_Table}.new");
	}
}
# Must be a collect run or something on one node
elsif ( $node_search ne "" ) {
	if ( $node_search eq $NMIS::nodeTable{$node_search}{node} ) {
		&runNodeStats($node_search);
	}
	else {
		print "Invalid node $node_search: No node of that name!\n";
	}
}
# Must be a collect run or something so do all the nodes
else {
	##############################################################
	# Processing Loop for all nodes
	##############################################################	
	$index = 0;
	
	#! multithreading
	#! currentthreads should be 0....
	if ($mthread)
	{
		#! shlock will lock the shared memory var $currentThreads so that no other process will write something in it.
		#! all other processes will be queued! So nothing will get lost.
#!		$handle->shlock();
#!		$currentThreads=0;
		#! unlocks the var
#!		$handle->shunlock();
	}

	foreach $onenode ( keys (%NMIS::nodeTable) )
	{
		# on each loop verify if runstats has been running too
		# long, each pass should always be less then 30 minutes
		if ( time - $start_time > $run_time and $type ne "update" ) {
			$start_time = returnDateStamp($start_time);
			&errorQuit ("$0 has been running for more then $run_time seconds exiting....\nStart Time $start_time. Processed $index nodes. Current node=$onenode.");
		}
		
		# This will allow debugging to be turned on for a
		# specific node where there is a problem
		if ( $onenode eq "$debug_watch" ) { $debug = "true"; }
		else { $debug = $debug_global; }

		undef %reach;
		
		#! One thread for each node until maxThreads is reached.		
		#! This loop is entered only if the commandlinevariable mthread=true is used!
		if ($mthread)
		{
			#! incrementing the currentThread value for each thread which is startet. This incrementation will be done in the
			#! father process because it can take a time until the child is borne and running
#!			$handle->shlock();
#!			$currentThreads++;
#!			$handle->shunlock();
			
			if ($mthreadDebug)
			{
      	print "FATHER-> running in multithreading modus\n";
			}
			#! starting new child process
			if ($pid=fork)
			{	
				#! this will be run only by the father process
				if ($mthreadDebug)
				{
					$currentThreads=Proc::Queue::running_now();
					print "FATHER-> starting a thread for node $onenode\n";
					print "FATHER-> currently there are $currentThreads threads running\n";
				}
			}
			#! this will be run only by the child
			else
			{
				if ($mthreadDebug)
				{				
					print "CHILD $$-> I am a CHILD with the PID $$ \n";
					print "CHILD $$-> My father's PID is ";print getppid();print "\n";
				}

				&runNodeStats($onenode);

			  #! all the work in this thread is done
				#! now this child will die. But before it must decrement the currentThreads var
				if ($mthreadDebug)
				{				
					$currentThreads=Proc::Queue::running_now();
					print "CHILD $$-> currently there are $currentThreads running\n";
				}

#!				$handle->shlock();
#!				$currentThreads=$currentThreads-1;
#!				$handle->shunlock();

				if ($mthreadDebug)
				{				
					$currentThreads=Proc::Queue::running_now();
					print "CHILD $$-> now there are $currentThreads running\n";
					print "CHILD $$-> I WiLl DiE nOw aRgHhHhHhHhH\n";
				}
				
				#! killing child
				exit 0;
			}
		}

		#! will be run if mthread is false (no multithreading)
		else
		{
			&runNodeStats($onenode);
		}

		#! Check how much threads are running.
#!		if ($mthread)
#!		{	
#!			if ($currentThreads == $maxThreads)
#!			{
#!				if ($mthreadDebug)
#!				{
#!					print "FATHER-> maximum number of concurrent threads is reached!\n";
#!					print "FATHER-> there are $currentThreads  Threads\n";
#!				}
				
				#! wait until one or more childs are dead....
#!				while ($currentThreads >= $maxThreads)
#!				{
#!					if ($mthreadDebug)
#!					{
#!						print "FATHER-> must wait until one or more childs are dead. Currently there are $currentThreads Threads\n";
#!					}
					
					#! sleep a while. is good for the cpu....
#!					select(undef, undef, undef, 0.3);
#!				}

#!				if ($mthreadDebug)
#!				{
#!					print "FATHER-> One or more Childs are dead now. So I can fork some new ones.\n";
#!					print "FATHER-> currently there are $currentThreads Threads\n";
#!				}
#!			}
#!		} #if mthread
	} # foreach $onenode
}

if ($mthreadDebug)
{
	print "FATHER-> All Threads are running.\n";
	$currentThreads=Proc::Queue::running_now();
	print "FATHER-> Must wait now for $currentThreads childs to finish.\n";
}

# Wait for all childs
while (wait != -1) {};

if ($mthreadDebug)
{
	print "FATHER-> All Threads are dead.\n";
	print "FATHER-> Will now continue without multithreading.\n";
}

#! removing all shared memory created by nmis
#!IPC::Shareable->clean_up();

if ($debug) {
	$endTime = sprintf( "%.2f", Time::HiRes::tv_interval([$startTime]));
	print "\n".returnTime ." Polled $nodecount nodes current runtime $endTime seconds.\n";
};

# Couple of post processing things.
if ( $type eq "collect" ) { &runEscalate; }
elsif ( $type eq "update" ) { &createInterfaceInfo; }

if ($debug) { 
	$endTime = sprintf( "%.2f", Time::HiRes::tv_interval([$startTime]));
	print "\n".returnTime ." End of $0 Ran for $endTime seconds.  BYE!\n\n";
};

exit 0;

sub runNodeStats {
